from . import scndup
from ..catalog import ScanCatalog, StationCatalog
from ..util import f2str

import schedlib as s

import numpy as np

in_scan = 0
def optupt(last_scan_index, k_scan, scan_index):
    global in_scan

    done = False
    keep = True
    adjust = False
    
    stations = StationCatalog().used(use_direct_access=True)
    scans = ScanCatalog().direct_access_entries
    
    if k_scan == 1:
        s.wlog(0, "OPTUPT:  OPTMODE=UPTIME was specified.  Station files "
               "(sch, crd, ")
        s.wlog(0, "         vex, and drudg) will not be written.")

        in_scan = 1
        scans[scan_index - 1].startj = scans[0].startj
        if s.schcon.opdur <= 0:
            s.errlog(" OPTCUPT: For OPTMODE=UPTIME, OPDUR must be given.")

        s.wlog(0, "         There are {} input scans with sources: ".format(
            s.schn1.nscans))
        sources = set()
        for j_scan, scan in enumerate(scans[:s.schn1.nscans]):
            s.wlog(0, "     {:>5d}: {}".format(j_scan + 1, scans[j_scan].scnsrc))
            sources.add(scans[j_scan].scnsrc)

        if len(sources) < s.schn1.nscans:
            s.wlog(1, "       There were duplicate sources. ")
            s.wlog(1, "       This is probably not what you wanted for "
                   "OPTMODE=UPTIME.")
    else:
        scans[scan_index - 1].startj = scans[scan_index - 2].stopj + \
                                       scans[in_scan - 1].gap

    scans[scan_index - 1].stopj = scans[scan_index - 1].startj + \
                                  scans[in_scan - 1].dur
    if scans[scan_index - 1].stopj > scans[0].startj + s.schcon.opdur:
        if in_scan >= s.schn1.nscans:
            done = True
            keep = False
        else:
            in_scan += 1
            scans[scan_index - 1].startj = scans[0].startj
            scans[scan_index - 1].stopj = scans[scan_index - 1].startj + \
                                          scans[in_scan - 1].dur
            last_scan_index.fill(0)

    if not done:
        scndup(scan_index - 1, in_scan - 1, False, "OPTUPT")

        for i, station in enumerate(stations):
            if station.stascn[scan_index - 1]:
                s.stageo(scan_index, i + 1, scans[scan_index - 1].startj, 0, 
                         "OPTUPT")
                station.stascn[scan_index - 1] = \
                    (f2str(station.up1[scan_index - 1]) == "") and \
                    (f2str(station.up2[scan_index - 1]) == "") and \
                    ((station.el1[scan_index - 1] + 
                      station.el2[scan_index - 1]) / 2
                     > scans[scan_index - 1].opminel)

    return last_scan_index, adjust, keep, done



#! /usr/bin/env python3

from collections import Counter
import os.path
import re
import sys

def getAntennas (filename, stopPattern=None, startPattern=None, useCounter=False):

    if useCounter:
        antennas = Counter()
        def addOne (element, collection):
            if element in collection:
                collection[element]+=1
            else:
                collection[element]=1 
    else:
        antennas = set() if not useCounter else Counter()
        def addOne (element, collection):
            collection.add(element)

    skipping = startPattern is not None
    if skipping:
        print (f"--> Skipping in {filename}")

    if stopPattern is not None:
        print (f"--> Reading until stop pattern in {filename}")

    with open (filename) as data:

        for line in data:

            if skipping:
                skipping = line.rstrip() != startPattern
                if not skipping:
                    print (f"*** Ending skip for {filename=} at '{line}'")
                continue

            if stopPattern is not None and stopPattern == line.rstrip():
                print (f"Stopping at line: '{line}'")
                break

            m = re.match ("^ .*DBNAME=(?P<antenna>[^ ]+).*$", line)

            if m:
                addOne (m.group("antenna"), antennas)

    return antennas


def doIt (originalFile="locations-original.dat", newFile="locations.dat"):

    # Read in the old and new locations files

    print (f"Original locations file: {originalFile}")
    print (f"New locations file: {newFile}")


    originalAntennas = getAntennas (originalFile)
    newAntennas = getAntennas (newFile)

    # Determine if there are any missing files

    missing = originalAntennas.difference (newAntennas)

    print ("Comparing new file to original file:")

    if len(missing) != 0:
        print (f"--> New file is missing: {missing}")
    else:
        print ("--> No missing antennas")

    newAntennas = getAntennas ("locations.dat", useCounter=True)
    
    duplicates = [(k,v) for k,v in newAntennas.items() if v > 1]

    print ("\nChecking new file for duplicates:")
    if len (duplicates) == 0:
        print ("--> No duplicates.")
    else:
        print (f"--> Duplicated antennas: {duplicates=}")


    

    # newAntennas = getAntennas ("locations.geodesy.dat")
    # replacedAntennas = getAntennas ("locations-original.dat", 
    #                                 stopPattern="!   From file $SCHED/catalogs/Master_NRAO/locations_VLA.dat")

    # conflictingAntennas =  getAntennas ("locations-original.dat", 
    #                                     startPattern="!   From file $SCHED/catalogs/Master_NRAO/locations_VLA.dat")

    # print (f"*** {len(originalAntennas)} {sorted(list(originalAntennas))=}")
    # print (f"*** {len(newAntennas)} {sorted(list(newAntennas))=}")
    # print (f"*** {len(replacedAntennas)} {sorted(list(replacedAntennas))=}")
    # print (f"*** {len(conflictingAntennas)} {sorted(list(conflictingAntennas))=}")

    # s1 = originalAntennas.difference (replacedAntennas) 
    # updatedAntennas = s1.union(newAntennas)

    # missingAntennas = originalAntennas.difference (updatedAntennas)

    # if len(missingAntennas) != 0:
    #     print (f"*** {missingAntennas=}")
    # else:
    #     print ("--> No missing antennas")

    # conflict = conflictingAntennas.intersection (newAntennas)

    # if len (conflict) != 0:
    #     print (f"*** {conflict=}")
    # else:
    #     print ("--> No conflicting antennas")

    


    
if __name__ == "__main__":


    if "--help" in sys.argv or len (sys.argv) < 3:
        thisFile = os.path.basename (sys.argv[0])
        print (f"*** usage: {thisFile} oldFilename newFilename\n")
        print ("Using old locations file as a reference checks for ")
        print ("antennas missing in the new file. Also checks for duplicate")
        print ("antennas in the new file.")
        sys.exit (1)

    doIt(sys.argv[1], sys.argv[2])



from .. import util

import schedlib as s

def getcor(values):
    if s.schcon.debug:
        s.wlog(0, "GETCOR starting.")

    miscor = False

    correl = values["correl"]
    s.schsco.correl = util.resize_string(correl, s.schsco.correl.itemsize, 
                                         "correl")
    if not s.schn1.notape and correl not in ("SOCORRO",
                                             "VLBADIFX",
                                             "VLBA",
                                             "HAYSTACK",
                                             "BONN",
                                             "JIVE",
                                             "WASHINGTON",
                                             "USNO",
                                             "JPL",
                                             "BOLOGNA",
                                             "MITAKA",
                                             "PENTICTON",
                                             "LBA",
                                             "FXCORR",
                                             "ASC",
                                             "OTHER"):
        s.wlog(1, " ** WARNING: {} is not a recognized correlator.".format(
            correl))
        s.wlog(1, "     Recognized correlators are: SOCORRO, VLBADIFX, "
               "VLBA, HAYSTACK, BONN, JIVE,")
        s.wlog(1, "       USNO, JPL, BOLOGNA, MITAKA, PENTICTON, LBA, "
               "FXCORR, ASC, and OTHER")
        miscor = True

    s.schco.coravg = values["coravg"][0]
    s.schco.caexact = (values["coravg"][1].upper() == "EXACT")
    s.schco.corav2 = values["coravg2"][0]
    s.schco.caexact2 = (values["coravg2"][1].upper() == "EXACT")

    s.schco.corchan = values["corchan"][0]
    s.schco.corfft = values["corchan"][1] if values["corchan"][1] != 0 \
                     else max(128, s.schco.corchan)
    s.schco.cornant = values["cornant"]
    s.schco.corpol = (values["corpol"] == "ON")
    
    s.schsco.corwtfn = util.resize_string(values["corwtfn"], 
                                          s.schsco.corwtfn.itemsize, "corwtfn")
    s.schsco.cortape = util.resize_string(values["cortape"], 
                                          s.schsco.cortape.itemsize, "cortape")
    s.schsco.cordfmt = util.resize_string(values["cordfmt"], 
                                          s.schsco.cordfmt.itemsize, "cordfmt")
    s.schsco.corsrcs = util.resize_string(values["corsrcs"], 
                                          s.schsco.corsrcs.itemsize, "corsrcs")
    for i in range(4):
        corship = "corship" + str(i+1)
        s.schsco.corship[i] = util.resize_string(
            values[corship], s.schsco.corship.itemsize, corship)
        cornote = "cornote" + str(i+1)
        s.schsco.cornote[i] = util.resize_string(
            values[cornote], s.schsco.cornote.itemsize, cornote)

    if not s.schn1.notape:
        s.chkcor()
        
        if correl in ("SOCORRO",
                      "VLBADIFX",
                      "VLBA",
                      "FXCORR"):
            s.socdef(values["corpol"])
        else:
            s.cordef(values["corpol"])
    
    if correl in ("SOCORRO",
                  "VLBADIFX",
                  "VLBA"):
        correlator_text = ("  Correlator (Required if recording):          {}"
                           "    (VLBA DiFX software correlator)".format(correl))
    elif miscor:
        correlator_text = ("  Correlator (Required if recording):          {}"
                           "    (Unknown correlator)".format(correl))
    else:
        correlator_text = ("  Correlator (Required if recording):          {}".\
                           format(correl))
    
    if len(values["corsrcs"]) <= 13:
        sources_text1 = "  Source positions from:             {}".format(
            values["corsrcs"])
        sources_text2 = ""
    else:
        sources_text1 = "  Source positions from:"
        sources_text2 = "  {}".format(values["corsrcs"])

    for index, line in enumerate((
            "CORRELATION REQUESTS (Defaults in parentheses): ",
            correlator_text,
            "  Correlator average time (2 sec):            {:8.3f} sec.".format(
                float(s.schco.coravg)),
            "    Alternate average time (for spacecraft).  {:8.3f} sec.".format(
                float(s.schco.corav2)),
            "  Output spectral channels per baseband (16):    {:8d}".format(
                int(s.schco.corchan)),
            "     Correlator FFT size (128):                  {:8d}".format(
                int(s.schco.corfft)),
            "  Number of antennas to be correlated:           {:8d}".format(
                int(s.schco.cornant)),
            "  Polarization (ON):                           {}".format(
                values["corpol"]),
            "  Correlator weighting function (UNIFORM):     {}".format(
                values["corwtfn"]),
            "  Distribution tape (DAT):                     {}".format(
                values["cortape"]),
            "  Distribution format (FITS):                  {}".format(
                values["cordfmt"]),
            sources_text1,
            sources_text2,
            "  Shipping address for correlator output: ",
            "      {}".format(values["corship1"]),
            "      {}".format(values["corship2"]),
            "      {}".format(values["corship3"]),
            "      {}".format(values["corship4"]),
            "  Correlator Notes: {}".format(values["cornote1"]),
            "      {}".format(values["cornote2"]),
            "      {}".format(values["cornote3"]),
            "      {}".format(values["cornote4"]),
            " "
    )):
        s.schsco.corstuff[index] = util.resize_string(
            line, s.schsco.corstuff.itemsize, "corstuff")



from .catalog import Catalog

import schedlib as s

class PeakCatalog(Catalog):
    """
    INTEGER      NPKSTA(MPKGRP), NPKSRC(MPKGRP)
    INTEGER      PKLSET(MPKGRP), PKLSETL(MPKGRP)
    CHARACTER    PKSRC(MPKSRC,MPKGRP)*12
    INTEGER      PKSRNUM(MPKSRC,MPKGRP)
C     Looking at the code, it seems like this should be PKMINEL(MPKGRP)
    REAL         PKMINEL(MPKSTA)
    CHARACTER    PKLINES(MPKGRP)*8
    CHARACTER    PSETFILE(MPKGRP)*80
    CHARACTER    PLSETFIL(MPKGRP)*80
    CHARACTER    PVLAMODE(MPKGRP)*2
    CHARACTER    PKSTA(MPKSTA,MPKGRP)*8
    DOUBLE PRECISION  PKDWELL(MPKGRP), PKMINFQ(MPKGRP)
    """

    maxpeak = s.schpeakn.npksta.shape[0]
    block_items = {
        s.schpeakn: [
            "npksta",
            "npksrc",
            "pklset",
            "pklsetl",
            "pksrnum",
            "pkminel",
            "pkdwell",
            "pkminfq"
        ],
        s.schpeakc: [
            "pklines",
            "psetfile",
            "plsetfil",
            "pvlamode",
            "pksta",
            "pksrc"
        ]
    }
        
    def __init__(self):
        super().__init__(self.maxpeak, self.block_items)

    def write(self, indices=None):
        if indices is None:
            entries = self.entries
        else:
            entries = [self.entries[i] for i in indices]
        for entry in entries:
            entry.npksrc = len(entry.pksrc)
            entry.npksta = len(entry.pksta)
        return super().write(indices)

    def adjust_lengths(self, entries):
        for entry in entries:
            entry.pksrc = entry.pksrc[:entry.npksrc]
            entry.pksta = entry.pksta[:entry.npksta]

    def scheduled_slice(self):
        return slice(s.schpeakn.npkgrp)



from . import optnone, optskd, optcells, optupt, opthas, opthiel, addpeak, addgeo
from ..catalog import ScanCatalog, StationCatalog, SourceCatalog, \
    PhaseCenterCatalog, SetupCatalog, PeakCatalog, SetupFileCatalog
from ..util import f2str

import schedlib as s

import numpy as np

import itertools

def schopt():
    if s.schcon.debug:
        s.wlog(0, "SCHOPT: Starting.")

    scan_catalog = ScanCatalog()
    scans = scan_catalog.direct_access_entries
    
    station_catalog = StationCatalog()
    stations = station_catalog.used(use_direct_access=True)

    source_catalog = SourceCatalog()
    sources = source_catalog.read()
    
    pc_catalog = PhaseCenterCatalog()
    phase_centers = pc_catalog.read()

    setups = SetupCatalog().read()
    peak_catalog = PeakCatalog() 
    peak_groups = peak_catalog.read()

    # content of setup file catalog used in geochk (->addgeo->geomake->geochk)
    # call here once
    SetupFileCatalog().read()

    for scan in scans:
        scan.origen = 1

    optmode = f2str(s.schsco.optmode)
    if (optmode not in {"NONE", "SCANS"}) or \
       s.schcon.autopeak or s.schsou.anygeo:
        # get the int value of nscans,
        # else (increment) operations affect nscans directly,
        # since it's a dimensionless np.array
        scan_index = int(s.schn1.nscans) 
        s.schn1.scan1 = scan_index + 1
    else:
        scan_index = 0
        s.schn1.scan1 = 1
    
    s.wlog(0, "SCHOPT:  First output scan will be number {}".format(
        s.schn1.scan1))

    done = False
    peak_opt = 0
    geo_opt = 0
    k_scan = 0
    last_scan_index = np.zeros(shape=(station_catalog.maxsta,), dtype=int)
    last_s_scan_index = np.zeros(shape=(station_catalog.maxsta,), dtype=int)
    while not done:
        scan_index += 1
        if scan_index > scan_catalog.maxscan:
            s.errlog("SCHOPT:  Trying to generate too many scans. Max: {}".\
                     format(scan_catalog.maxscan))
        
        scan = scans[scan_index - 1]
        if (peak_opt != 0) or (geo_opt != 0):
            keep = True
        else:
            k_scan += 1
            assert((k_scan == scan_index) or (scan_index >= s.schn1.nscans))
            if optmode == "NONE":
                adjust, keep, done = optnone(k_scan, scan_index)
            elif optmode == "SCANS":
                adjust, keep, done = optskd(last_scan_index, k_scan, scan_index)
            elif optmode == "CELLS":
                adjust, keep, done = optcells(
                    last_scan_index, k_scan, scan_index)
                scan.origen = 2
            elif optmode == "CSUB":
                s.errlog("SCHOPT: OPTMODE CSUB is not supported in pySCHED")
            elif optmode == "UPTIME":
                last_scan_index, adjust, keep, done = optupt(
                    last_scan_index, k_scan, scan_index)
                scan.origen = 2
            elif optmode == "HAS":
                adjust, keep, done = opthas(last_scan_index, k_scan, scan_index)
                scan.origen = 2
            elif optmode == "HIGHEL":
                k_scan, adjust, keep, done = opthiel(
                    last_scan_index, k_scan, scan_index)
                scan.origen = 2
            else:
                s.errlog("SCHOPT: Invalid OPTMODE: {}".format(optmode))

            if keep and (not done):
                s.opttim(last_scan_index, last_s_scan_index, scan_index, 
                    adjust, False, False)
                if (s.schcon.opdur != 0) and \
                   (scan.stopj > 
                    scans[s.schn1.scan1 - 1].startj + s.schcon.opdur) and \
                   (optmode != "HAS"):
                    done = True

                n_good = s.scngeo(last_scan_index, scan_index)
            
        insert_adjust = adjust
        if keep and (not done):
            if (peak_opt == 0) and ((scan.geolen > 0) or (geo_opt >= 1)):
                insert_adjust = False
                geo_opt, keep = addgeo(last_scan_index, scan_index, geo_opt, 
                                       scans, stations)

            peak_opt, insert_adjust = addpeak(
                last_scan_index, scans, stations, setups, peak_groups,
                peak_opt, insert_adjust, scan_index)
            peak_catalog.write()
            keep = s.makeptg(last_scan_index, scan_index, keep)

        if keep and (not done):
            s.opttim(last_scan_index, last_s_scan_index, scan_index, 
                insert_adjust, False, True)
            n_good = s.scngeo(last_scan_index, scan_index)
            n_good = s.autodown(last_scan_index, scan_index)
            
            keep = keep and ((n_good >= scan.opmian) or 
                             (optmode in {"CSUB", "HAS"}) or 
                             (scan.point >= 0))
            if not keep:
                for station in stations:
                    station.stascn[scan_index - 1] = False
            else:
                s.settps(scan_index, last_scan_index)
                
                for station_index, station in enumerate(stations, 1):
                    if station.stascn[scan_index - 1] and s.schn1.vlbitp and \
                       (not s.schcon.noset) and station.usedisk:
                        s.diskpos(scan_index, station_index, last_scan_index)
                
                if (optmode not in {"NONE", "UPTIME"}) or s.schsou.anygeo:
                    s.optsch(scan_index)

                source = sources[scan.srcnum - 1]
                source.sused = True
                if not scan.norec:
                    source.usedrec = True
                if scan.idopsrc != 0:
                    sources[scan.idopsrc - 1].sused = True
                if scan.ivlaphs != 0:
                    sources[scan.ivlaphs - 1].sused = True
                if scan.icent != 0:
                    source.usedcent = True
                    for source_index in phase_centers[scan.icent - 1].ctrsrci:
                        sources[source_index - 1].usedphs = True
        
            for station_index, station in enumerate(stations):
                if station.stascn[scan_index - 1]:
                    last_scan_index[station_index] = scan_index
                    if scan.origen < 4:
                        last_s_scan_index[station_index] = scan_index

            s.schn1.scanl = scan_index

    if s.schcon.debug:
        s.wlog(0, "SCHOPT: 9.")

    if s.schn1.scanl == 0:
        s.wlog(1, "SCHOPT: Did not schedule any scans")
        s.errlog(" Abort")

    source_catalog.write()
    s.accsrc(False)
    gotall = s.srcflg()

    if not gotall:
        s.errlog("SCHOPT: Not all sources found; programming problem.")

    s.getpairs()

    year, day1, start = s.timej(scans[s.schn1.scan1 - 1].startj)
    year, day2, stop = s.timej(scans[s.schn1.scanl - 1].stopj)
    
    time1 = f2str(s.tformwrp(start, "T", 0, 2, 2, "::@"))
    time2 = f2str(s.tformwrp(stop, "T", 0, 2, 2, "::@"))
    s.wlog(0, "SCHOPT:  There will be {number} output scans ({first} - {last}) "
           "from {day1}/{time1} to {day2}/{time2}".format(
               number=s.schn1.scanl - s.schn1.scan1 + 1,
               first=s.schn1.scan1, last=s.schn1.scanl,
               day1=day1, day2=day2, time1=time1, time2=time2))

    s.schtim()
    s.sch24()

    if s.schcon.debug:
        s.wlog(0, "SCHOPT: Done.")



from .catalog import Catalog

import schedlib as s

import numpy as np

class SourceCatalog(Catalog):
    """
    CHARACTER        SOURCE(MALIAS,MAXSRC)*12, SOUR8(MAXSRC)*8
    CHARACTER        CALCODE(MAXSRC)*1, REMARK(MAXSRC)*80
    DOUBLE PRECISION RACAT(MAXSRC),  DECCAT(MAXSRC)
    DOUBLE PRECISION RA1950(MAXSRC), D1950(MAXSRC)
    DOUBLE PRECISION RA2000(MAXSRC), D2000(MAXSRC)
    DOUBLE PRECISION RAP(MAXSRC), DECP(MAXSRC), PMTIME(MAXSRC)
    DOUBLE PRECISION EPOCHT(MAXSRC)
    DOUBLE PRECISION PMRA(MAXSRC), PMDEC(MAXSRC), PARALAX(MAXSRC)
    CHARACTER*1      C1950(MAXSRC), C2000(MAXSRC), CDATE(MAXSRC)
    REAL             RAERR(MAXSRC), DECERR(MAXSRC)
    REAL             DRA(MAXSRC), DDEC(MAXSRC), VLSR(MAXLCH,MAXSRC)
    CHARACTER        WHICHCAT(MAXSRC)*1 
    CHARACTER        VELREF(MAXSRC)*1, VELDEF(MAXSRC)*1
    LOGICAL          SUSED(MAXSRC), DOPPED(MAXSRC)
    LOGICAL          USEDREC(MAXSRC), USEDPHS(MAXSRC)
    LOGICAL          USEDCENT(MAXSRC)
    INTEGER          SRLSTN(MAXSRC), DIDNDOP(MAXSRC), SATN(MAXSRC)
    REAL             SUNDIS(MAXSRC)
    LOGICAL          PLANET(MAXSRC), SATEL(MAXSRC)
    CHARACTER        CSUSED(MALIAS,MAXSRC)*1
    INTEGER          SRCATN(MAXSRC)
    """
    maxsource = s.schn1.srcatn.shape[0]
    block_items = {
        s.schsou: [
            "racat",
            "deccat",
            "ra1950",
            "d1950",
            "ra2000",
            "d2000",
            "rap",
            "decp",
            "pmtime",
            "epocht",
            "pmra",
            "pmdec",
            "paralax",
            "raerr",
            "decerr",
            "dra",
            "ddec",
            "vlsr",
            "sused",
            "dopped",
            "usedrec",
            "usedphs",
            "usedcent",
            "srlstn",
            "didndop",
            "satn",
            "sundis",
            "planet",
            "satel"
        ],
        s.schn1: [
            "srcatn"
        ],
        s.schcsc: [
            "source",
            "sour8",
            "calcode",
            "remark",
            "c1950",
            "c2000",
            "cdate",
            "whichcat",
            "velref",
            "veldef",
            "csused"
        ]
    }

    def __init__(self):
        super().__init__(self.maxsource, self.block_items)

    def scheduled_slice(self):
        return slice(s.schsou.msrc)

    def used(self):
        return [source for source in self.entries if source.sused]

    def set_aliases(self):
        """
        Pre: self has initialized entries
        """
        for source in self.scheduled():
            source.aliases = source.source[np.in1d(source.csused, ["+", "*"])]
        




from .scan_catalog import ScanCatalog
from .station_catalog import StationCatalog
from .frequency_catalog import FrequencyCatalog
from .peak_catalog import PeakCatalog
from .line_rest_frequency_catalog import LineRestFrequencyCatalog
from .source_catalog import SourceCatalog
from .satellite_catalog import SatelliteCatalog
from .phase_center_catalog import PhaseCenterCatalog
from .setup_catalog import SetupCatalog
from .setup_file_catalog import SetupFileCatalog
from .source_plot_catalog import SourcePlotCatalog



from .catalog import Catalog

import schedlib as s

class SatelliteCatalog(Catalog):
    """
    INTEGER          SATNUM(MAXSAT)
    CHARACTER        SATNAME(MAXSAT)*12
    CHARACTER        SATFILE(MAXSAT)*128
    CHARACTER        TLEFILE(MAXSAT)*128
    CHARACTER        KERFILE(MAXSAT)*128
    """
    maxsatellite = s.schsat.satnum.shape[0]
    block_items = {
        s.schsat: [
            "satnum"
        ],
        s.schcat: [
            "satname", 
            "satfile",
            "tlefile",
            "kerfile"
        ]
    }

    def __init__(self):
        super().__init__(self.maxsatellite, self.block_items)




from . import geomake, gmkscn
from .parameter import max_seg, secpday
from ..util import f2str
from ..catalog import StationCatalog, ScanCatalog

import schedlib as s

import numpy as np

geo_stascn = np.full(fill_value=False, shape=(max_seg, s.schn2a.stascn.shape[1]))
geo_startj = np.empty(dtype=float, shape=(max_seg,))

n_seg = 0

def addgeo(last_scan_index, scan_index, geo_opt, scans, stations):
    global j_scan, n_seg, seg_sources, geo_stascn, geo_startj

    if s.schcon.debug:
        s.wlog(0, "ADDGEO starting.")

    if geo_opt == 0:
        j_scan = scans[scan_index - 1].geoiscn
        seg_sources, n_seg, geo_stascn, geo_startj = geomake(
            last_scan_index, j_scan, scan_index, n_seg, geo_stascn, geo_startj,
            stations, scans)
        geo_opt = n_seg

    if geo_opt != 0:
        seg_index = n_seg - geo_opt
        scan_stascn = geo_stascn[seg_index, :]
        approx_time = geo_startj[seg_index]
        seg_source_index = seg_sources[seg_index]
        
        n_good, ok_sta, scan_stascn = gmkscn(
            last_scan_index, scan_index, j_scan, 
            s.schsou.geosrci[seg_source_index - 1], 
            s.schcsc.geosrc[seg_source_index - 1], 
            approx_time, scans[j_scan - 1].opminel, 0, scan_stascn, "FORCE")

        geo_opt -= 1
        scan = scans[scan_index - 1]
        scan.origen = 3
        if s.schsou.geoprt >= 0:
            if scan_index == s.schn1.scan1:
                s_gap = 0.
            else:
                s_gap = (scan.startj - scans[scan_index - 2].stopj) * secpday

            msg = "{:4d} {:8.0f} {:<12}".format(
                seg_source_index, s_gap, 
                f2str(s.schcsc.geosrc[seg_source_index - 1]))
            msg += " ".join(" {:4.0f} ".format(s.el1[scan_index - 1]) 
                            if s.stascn[scan_index - 1]
                            else "({:4.0f})".format(s.el1[scan_index - 1]) 
                            for s in stations[:20])
            s.wlog(0, msg)

    return geo_opt, True



"""
Module to update the toolbar of matplotlib plots to work around a few 
issues specific to pySCHED (no proper date editor, as the get_[xy]lim function 
returns a float, edit all 3 axis at the same time, LST limits need their own 
widget, ...).

The main function is adjust. This will work on the toolbar after it is created. 
It will remove a widget and its action and insert a new one. A similar approach 
is taken with the the resulting popup dialog (remove and insert). The 
alternative is to subclass the toolbar and dialog classes, but that would 
cause a lot of code copy-pasting. Even with this approach there is some 
copy-pasting, therefor the matplotlib license:

License Agreement (MIT License)
------------------------------------------

Copyright (c) 2009 Pierre Raybaut

Permission is hereby granted, free of charge, to any person
obtaining a copy of this software and associated documentation
files (the "Software"), to deal in the Software without
restriction, including without limitation the rights to use,
copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
"""
from .plot_util import shut_up_mpl

import matplotlib
import matplotlib.backends.qt_editor.figureoptions as figureoptions
import matplotlib.dates
from matplotlib import cbook
import six
from matplotlib.colors import rgb2hex
from matplotlib.colors import colorConverter
from matplotlib.backend_tools import ToolBase
from matplotlib.backends.qt_compat import QtGui

# hack around bug in formlayout, set an environment variable to force use of Qt5
import os; os.environ["QT_API"] = "pyqt5"
import formlayout

from distutils.version import LooseVersion
from PyQt5.Qt import PYQT_VERSION_STR
if LooseVersion(PYQT_VERSION_STR) < LooseVersion("5.8"):
    raise ImportError("PyQt version 5.8 or higher is required")
from PyQt5.QtWidgets import QWidget, QTimeEdit, QMessageBox, QHBoxLayout, \
    QSpinBox, QAction, QInputDialog
from PyQt5.QtCore import QDateTime

from datetime import datetime, timedelta
from pathlib import Path

# many API changes from version 1.1 -> 1.2 in formlayout.FormDialog,
# wrap the call to the constructor
if LooseVersion(formlayout.__version__) < LooseVersion("1.2"):
    def FormDialogWrapper(data, title, icon, apply):
        return formlayout.FormDialog(
            data, title=title, icon=icon, apply=apply)
else:
    def FormDialogWrapper(data, title, icon, apply):
        def apply_wrapper(results, widgets):
            apply(results)
        return formlayout.FormDialog(
            data, title, "", icon, None, apply_wrapper, True, True, "list",
            None, "form", False, None, None)

class LstWidget(QWidget):
    def __init__(self, value, lst_base_date, parent=None):
        super().__init__(parent)
        layout = QHBoxLayout(self)
        self.lst_base_date = lst_base_date
        # offset to lst base edit
        self.days = QSpinBox(self)
        self.days.setMinimum(-10) # somewhat arbitrary value
        self.days.setValue((value.date() - lst_base_date).days)
        layout.addWidget(self.days)
        self.time = QTimeEdit(self)
        self.time.setDisplayFormat("hh:mm:ss")
        self.time.setTime(value.time())
        layout.addWidget(self.time)

    def dateTime(self):
        # will be called by .get of the dialog, expecting a QDateTime return
        return QDateTime(datetime.combine(self.lst_base_date, 
                                          self.time.time().toPyTime()) +
                         timedelta(days=self.days.value()))

class CustomizeCurvesTool(ToolBase):
    default_keymap = "u"
    description = "Customize axes and curves"
    image = str(Path(matplotlib.get_data_path(),
                     "images", "qt4_editor_options.svg"))

    def __init__(self, *args, x_axis_type, **kwargs):
        with shut_up_mpl():
            super().__init__(*args, **kwargs)
        self.x_axis_type = x_axis_type

    def trigger(self, *args, **kwargs):
        all_axes = self.figure.get_axes()
        if not all_axes:
            QMessageBox.warning(
                self.figure.canvas.parent(),
                "Error", "There are no axes to edit.")
            return
        elif len(all_axes) == 1:
            axis = all_axes[0]
        elif self.x_axis_type in {"GST", "LST", "UT"}:
            # edit all axes at the same time, 
            # only makes sense if same data shape is edited
            axis = all_axes[-1]
        else:
            titles = [
                ax.get_label() or
                ax.get_title() or
                " - ".join(filter(None, [ax.get_xlabel(), ax.get_ylabel()])) or
                f"<anonymous {type(ax).__name__}>"
                for ax in all_axes]
            duplicate_titles = [
                title for title in titles if titles.count(title) > 1]
            for i, ax in enumerate(all_axes):
                if titles[i] in duplicate_titles:
                    titles[i] += f" (id: {id(ax):#x})"  # Deduplicate titles.
            item, ok = QInputDialog.getItem(
                self.figure.canvas.parent(), "Customize", "Select axes:",
                titles, 0, False)
            if not ok:
                return
            axis = all_axes[titles.index(item)]

        sep = (None, None)  # separator

        has_curve = len(axis.get_lines()) > 0

        # Get / General
        if self.x_axis_type in ["GST", "LST"]:
            xmin, xmax = (matplotlib.dates._from_ordinalf(l) 
                          for l in axis.get_xlim())
        elif self.x_axis_type == "UT":
            xmin, xmax = (matplotlib.dates._from_ordinalf(l) 
                          for l in axis.get_xlim())
        else:
            xmin, xmax = axis.get_xlim()
        ymin, ymax = axis.get_ylim()

        general = [('Title', axis.get_title()),
                   sep,
                   (None, "<b>X-Axis</b>"),
                   ('Min', xmin), 
                   ('Max', xmax),
                   ('Scale', [axis.get_xscale(), 'linear', 'log']),
                   sep,
                   (None, "<b>Y-Axis</b>"),
                   ('Min', ymin), 
                   ('Max', ymax),
                   ('Scale', [axis.get_yscale(), 'linear', 'log']),
                   sep,
                   ('(Re-)Generate automatic legend', False),
                   ]
        xmin_row = 3
        xmax_row = 4

        # Save the unit data
        xconverter = axis.xaxis.converter
        yconverter = axis.yaxis.converter
        xunits = axis.xaxis.get_units()
        yunits = axis.yaxis.get_units()
        xformatter = axis.xaxis.get_major_formatter()
        yformatter = axis.yaxis.get_major_formatter()

        if has_curve:
            # Get / Curves
            linedict = {}
            for line in axis.get_lines():
                label = line.get_label()
                if label == '_nolegend_':
                    continue
                linedict[label] = line
            curves = []
            linestyles = list(six.iteritems(figureoptions.LINESTYLES))
            drawstyles = list(six.iteritems(figureoptions.DRAWSTYLES))
            markers = list(six.iteritems(figureoptions.MARKERS))
            curvelabels = sorted(linedict.keys())
            for label in curvelabels:
                line = linedict[label]
                color = rgb2hex(colorConverter.to_rgb(line.get_color()))
                ec = rgb2hex(colorConverter.to_rgb(line.get_markeredgecolor()))
                fc = rgb2hex(colorConverter.to_rgb(line.get_markerfacecolor()))
                curvedata = [('Label', label),
                             sep,
                             (None, '<b>Line</b>'),
                             ('Line Style',
                              [line.get_linestyle()] + linestyles),
                             ('Draw Style',
                              [line.get_drawstyle()] + drawstyles),
                             ('Width', line.get_linewidth()),
                             ('Color', color),
                             sep,
                             (None, '<b>Marker</b>'),
                             ('Style', [line.get_marker()] + markers),
                             ('Size', line.get_markersize()),
                             ('Facecolor', fc),
                             ('Edgecolor', ec),
                             ]
                curves.append([curvedata, label, ""])

            # make sure that there is at least one displayed curve
            has_curve = bool(curves)

        datalist = [(general, "Axes", "")]
        if has_curve:
            datalist.append((curves, "Curves", ""))

        def apply_callback(data):
            """This function will be called to apply changes"""
            if has_curve:
                general, curves = data
            else:
                general, = data

            # Set / General
            title, xmin, xmax, xscale, ymin, ymax, yscale, \
                generate_legend = general

            axes_to_apply = [axis]
            if self.x_axis_type in {"GST", "LST", "UT"}:
                axes_to_apply = all_axes
            for apply_to_axis in axes_to_apply:
                apply_to_axis.set_xscale(xscale)
                apply_to_axis.set_yscale(yscale)
                apply_to_axis.set_title(title)
                apply_to_axis.set_xlim(xmin, xmax)
                apply_to_axis.set_ylim(ymin, ymax)

                # Restore the unit data
                apply_to_axis.xaxis.converter = xconverter
                apply_to_axis.yaxis.converter = yconverter
                apply_to_axis.xaxis.set_units(xunits)
                apply_to_axis.yaxis.set_units(yunits)
                apply_to_axis.xaxis.set_major_formatter(xformatter)
                apply_to_axis.yaxis.set_major_formatter(yformatter)
                apply_to_axis.xaxis._update_axisinfo()
                apply_to_axis.yaxis._update_axisinfo()

                if has_curve:
                    # Set / Curves
                    for index, curve in enumerate(curves):
                        line = linedict[curvelabels[index]]
                        label, linestyle, drawstyle, linewidth, color, \
                            marker, markersize, markerfacecolor, \
                            markeredgecolor = curve
                        line.set_label(label)
                        line.set_linestyle(linestyle)
                        line.set_drawstyle(drawstyle)
                        line.set_linewidth(linewidth)
                        line.set_color(color)
                        if marker != 'none':
                            line.set_marker(marker)
                            line.set_markersize(markersize)
                            line.set_markerfacecolor(markerfacecolor)
                            line.set_markeredgecolor(markeredgecolor)

                # re-generate legend, if checkbox is checked

                if generate_legend:
                    draggable = None
                    ncol = 1
                    if apply_to_axis.legend_ is not None:
                        old_legend = apply_to_axis.get_legend()
                        draggable = old_legend._draggable is not None
                        ncol = old_legend._ncol
                    new_legend = apply_to_axis.legend(ncol=ncol)
                    if new_legend:
                        new_legend.draggable(draggable)

            # Redraw
            figure = apply_to_axis.get_figure()
            figure.canvas.draw()

        dialog = FormDialogWrapper(
            datalist, 
            title="Figure options", 
            icon=QtGui.QIcon(self.image), 
            apply=apply_callback)
        axes_tab = dialog.formwidget.widgetlist[0]
        if LooseVersion(formlayout.__version__) >= LooseVersion("1.2"):
            # in version >= 1.2 the widgetlist elements are
            # (title, widget) tuples
            axes_tab = axes_tab[1]

        if self.x_axis_type in ["GST", "LST"]:
            layout = axes_tab.formlayout
            for (value, row) in ((xmin, xmin_row), (xmax, xmax_row)):
                layout.removeRow(row)
                widget = LstWidget(value, axis.xaxis.lst_base_date, axes_tab)
                layout.insertRow(row, general[row][0], widget)
                axes_tab.widgets[row] = widget
        elif self.x_axis_type == "UT":
            for row in (xmin_row, xmax_row):
                axes_tab.widgets[row].setDisplayFormat("yyyy-MM-dd hh:mm:ss")

        if dialog.exec_():
            apply_callback(dialog.get())


def adjust(figure, x_axis_type=None):
    """
    Adjust the toolbar of @figure to include a "Customize axis and curves"
    button.
    If x_axis_types is any of "GST", "LST" or "UT", the edit boxes for that
    axis will be in dates instead of floating point numbers.
    """
    manager = figure.canvas.manager
    manager.toolmanager.add_tool(
        "customize", CustomizeCurvesTool, x_axis_type=x_axis_type)
    manager.toolbar.add_tool("customize", "custom", 0)



from .catalog import Catalog

import schedlib as s

class SetupFileCatalog(Catalog):
    """
    DOUBLE PRECISION SFFREQ(MAXCHN,MAXSET)
                     SFFILT(MAXCHN,MAXSET)
    INTEGER          MSCHN(MAXSET)
    INTEGER          SAMEBBC(MAXCHN,MAXSET)
                     MINBBC(MAXSET)
    REAL             FSPEED(MAXSET)
    LOGICAL          OKXC(MAXSET)
    CHARACTER        SETFILE(MAXSET)*80
                     SFSIDE(MAXCHN,MAXSET)*1
    CHARACTER        SFPOL(MAXCHN,MAXSET)*3
    """
    
    max_setup_files = s.schsf.mschn.shape[0]
    block_items = {
        s.schsf: [
            "sffreq", 
            "sffilt", 
            "mschn", 
            "samebbc", 
            "minbbc", 
            "fspeed", 
            "okxc"
        ],
        s.schssf: [
            "setfile", 
            "sfside", 
            "sfpol"
        ]
    }

    def __init__(self):
        super().__init__(self.max_setup_files, self.block_items)

    def scheduled_slice(self):
        return slice(s.schsf.nsetf)

    @classmethod
    def extend_with(cls, filename, error_message):
        """
        Write a new catalog entry directly into the common blocks
        and return the (Fortran based) index of the entry
        """
        catalog = cls()
        catalog.read()
        # create an entry for the setup file if it is new
        index = next((index for index, entry in enumerate(catalog.scheduled())
                      if entry.setfile == filename), None)
        if (index is None) and (s.schsf.nsetf < catalog.max_setup_files):
            index = int(s.schsf.nsetf)
            catalog.entries[index].setfile = filename
            s.schsf.nsetf += 1
            catalog.write(range(index, int(s.schsf.nsetf)))
        elif (index is None):
            s.errlog(error_message)
        return index + 1



from .sched.parameter import secpday

import schedlib as s

# 22 mins of 2 Gbps recording (allow factor 1.008 for headers).
max_disk_unit = 22.*60. * 2.016 * (1024./1000.) / 8.
gap_seconds = 10
minimum_scan_seconds = 15
minimum_mode_change_gap_seconds = 40
maximum_tsys_gap_seconds = 15 * 60

def apply_tape_offset(scans, scan_offset, stations, setups):
    for scan_index, scan in enumerate(scans, scan_offset):
        stations_in_scan = [station for station in stations
                            if station.stascn[scan_index]]
        tape_starts = {station.tpstart[scan_index] 
                       for station in stations_in_scan}
        if len(tape_starts) == 0:
            continue
        if len(tape_starts) > 1:
            # -1 and +1: from FORTRAN index <-> python index
            min_pause = max(
                station.tpstart[scan_index]
                / setup[station.nsetup[scan_index]-1].speedup 
                for station in stations_in_scan)
            s.prtscn(scan_index + 1, "VXSCHK")
            s.errlog("Station tape starts differ, set MINPAUSE to {}s to "
                     "produce VEX file!".format(round(min_pause * secpday)))
        
        tape_start = tape_starts.pop()
        scan.startj -= tape_start
        
def is_field_system_station(station):
    return (station.control == "VEX") \
        and (not station.station.startswith("VLA"))

def check_recording_sizes(scans, scan_offset, stations):
    """
    Print warnings for large continuous recordings.
    """
    # gather the scan indices that trigger the warnings
    scan_warnings = set()
    for station in (s for s in stations
                    if is_field_system_station(s) 
                    and s.disk != "LBADR"
                    and s.usedisk):
        previous_scan_index = None
        previous_gbytes = 0
        for scan_index, scan in ((i, s) 
                                 for i, s in enumerate(scans, scan_offset)
                                 if station.stascn[i]):
            if previous_scan_index is not None:
                previous_scan = scans[previous_scan_index - scan_offset]
                if (scan.startj - previous_scan.stopj) * secpday > gap_seconds:
                    previous_gbytes = station.gbytes[previous_scan_index]
            
            gbytes = station.gbytes[scan_index]
            if gbytes - previous_gbytes > max_disk_unit:
                scan_warnings.add(scan_index)
            
            previous_scan_index = scan_index

    # print messages, do not print a new message for consecutive scans
    previous_scan_index = -42 # any value that triggers the check below
    for scan_index in sorted(scan_warnings):
        if scan_index != previous_scan_index + 1:
            s.wlog(0, "The scan detailed below has exceeded the limit for "
                   "continuous recording. Insert a gap before this scan, or "
                   "reduce its length if necessary:")
            s.prtscn(scan_index + 1, "VXSCH")
            s.wlog(0, " ")

        previous_scan_index = scan_index
        
    if len(scan_warnings) > 0:
        s.wrtmsg(1, "VXSCH", "warnbank")


def check_minimum_scan_duration(scans, scan_offset, stations):
    """
    Returns whether any scan failed the minimum duration check.
    """
    # only check certain stations, mimic SCHED here
    stations = [s for s in stations
                if is_field_system_station(s) 
                and s.disk != "LBADR"
                and s.usedisk]
    any_warning = False
    for scan_index, scan in enumerate(scans, scan_offset):
        if any(station.stascn[scan_index] for station in stations) \
           and ((scan.stopj - scan.startj) * secpday < minimum_scan_seconds):
            stations_string = ", ".join(station.station for station in stations
                                        if station.stascn[scan_index])
            s.prtscn(scan_index + 1, "VXSCHK")
            s.wlog(1, "WARNING: Scan length < {l}s for station(s) {s}.\n"
                   "         Currently FS supports minimal "
                   "scan length of {l}s".format(s=stations_string,
                                                l=minimum_scan_seconds))
            any_warning = True
    return any_warning

def check_scan_overlap(scans, scan_offset, stations):
    """
    Returns whether any scan failed the scan overlap check.
    """
    any_warning = False
    scan_pairs = zip(scans[:-1], scans[1:])
    for scan_index, (scan, next_scan) in enumerate(scan_pairs, scan_offset):
        if any(station.stascn[scan_index] and station.stascn[scan_index + 1]
               for station in stations) \
            and (next_scan.startj < scan.stopj):
            stations_string = ", ".join(station.station for station in stations
                                        if station.stascn[scan_index]
                                        and station.stascn[scan_index + 1])
            # print warning for next scan, +1 for next, +1 for FORTRAN indexing
            s.prtscn(scan_index + 2, "VXSCHK")
            s.wlog(1, "WARNING: Tape early start failed for stations(s) {}.".\
                   format(s=stations_string))
            s.wlog(1, "         Early tape starts only work if there are "
                   "sufficient gaps.")
            any_warning = True
    return any_warning

def check_mode_change(scans, scan_offset, stations, scan_mode):
    """
    scan_mode: {index in scans: mode name}
    Returns whether any scan failed the mode change gap check.
    """
    stations = [s for s in stations if is_field_system_station(s)]
    any_warning = False
    scan_pairs = zip(scans[:-1], scans[1:])
    for scan_index, (scan, next_scan) in enumerate(scan_pairs, scan_offset):
        if any(station.stascn[scan_index] and station.stascn[scan_index + 1]
               for station in stations) \
            and (scan_mode[scan_index - scan_offset] != 
                 scan_mode[scan_index - scan_offset + 1]) \
            and ((next_scan.startj - scan.stopj) * secpday 
                 < minimum_mode_change_gap_seconds):
            stations_string = ", ".join(station.station for station in stations
                                        if station.stascn[scan_index]
                                        and station.stascn[scan_index + 1])
            # print warning for next scan, +1 for next, +1 for FORTRAN indexing
            s.prtscn(scan_index + 2, "VXSCHK")
            s.wlog(1, "WARNING: Mode setup <= {l}s for station(s) {s}.\n"
                   "         FS Stations once needed {l}s for any mode "
                   "change, incl frequency shift.".format(
                       s=stations_string, l=minimum_mode_change_gap_seconds))
            s.wlog(1, "         This may no longer be true but the new value "
                   "is not yet clear.")
            any_warning = True
    return any_warning

def check_tsys(scans, scan_offset, stations, setups, frequency_setups):
    # TSCAL in the frequency setup will override the TSCAL of the station
    # (if present)

    tsys_warnings = []
    tsys_off_source_warnings = []
    for station in stations:
        last_stop = 0 # any MJD that will trigger a gap to the next scan
        n_tsys = 0
        n_tsys_on = 0
        last_tsys = scans[0].startj
        last_tsys_on_source = scans[0].startj
        warn_tsys_off_source = False
        max_tsys_gap = 0
        for scan_index, scan in enumerate(scans, scan_offset):
            if not station.stascn[scan_index]:
                continue
            scan_tscal = station.tscal
            # -1: FORTRAN -> python indexing
            setup = setups[station.nsetup[scan_index] - 1]
            # get the frequency setup index from the first channel,
            # this assumes they are all equal (as the SCHED code should enforce)
            if (setup.ifreqnum[0] >= 1):
                frequency_tscal = frequency_setups[setup.ifreqnum[0] -1].tscal
                if frequency_tscal is not None:
                    scan_tscal = frequency_tscal
            if scan_tscal == "CONT":
                # update last tsys measurement variables
                last_tsys = scan.stopj
                if station.tonsrc[scan_index] < scan.stopj:
                    last_tsys_on_source = scan.stopj
                continue
            
            # scan tscal is GAP, check interval between (on source) measurements
            if (scan.startj - last_stop) * secpday > gap_seconds:
                # a tsys measurement in this scan
                n_tsys += 1
                last_tsys = scan.startj
                on_source = round((station.tonsrc[scan_index] - scan.startj)
                                  * secpday)
                if on_source <= 0:
                    last_tsys_on_source = scan.startj
                    n_tsys_on += 1
                elif (scan.stopj - last_tsys_on_source) * secpday \
                     > maximum_tsys_gap_seconds:
                    warn_tsys_off_source = True

            tsys_gap = round((scan.stopj - last_tsys) * secpday)
            if tsys_gap > maximum_tsys_gap_seconds:
                max_tsys_gap = max(max_tsys_gap, tsys_gap)
                
            last_stop = scan.stopj

        if max_tsys_gap > 0:
            tsys_warnings.append(
                "{:<8} has {:>4} Tsys measurements. Maximum interval = {:>4} "
                "minutes.".format(station.station, n_tsys, 
                                  round(float(max_tsys_gap) / 60)))

        if warn_tsys_off_source:
            tsys_off_source_warnings.append(
                "{:<8}: only {:>4} out of {:>4} Tsys measurements "
                "are on-source".format(station.station, n_tsys_on, n_tsys))

    if len(tsys_warnings) > 0:
        for text in tsys_warnings:
            s.wlog(1, text)
        s.wlog(1, 
               "Tsys calibration at most MkIV stations is taken during "
               "every gap in recording,\n"
               "but these appear over {l} min apart for the stations listed "
               "above!\n"
               "This can be improved by inserting gaps at regular "
               "intervals.\n"
               "Note this is not an issue for Westerbork or Arecibo.".format(
                   l=round(maximum_tsys_gap_seconds/60)))
        s.wrtmsg(0, "VXSCH", "tsysgap")

    if len(tsys_off_source_warnings) > 0:
        for text in tsys_off_source_warnings:
            s.wlog(1, text)
        s.wlog(1, "Stations listed above are affected by slewing during "
               "Tsys calibration")
        s.wrtmsg(0, "VXSCH", "tsysoffsrc")

        
        



from ..catalog import SetupCatalog, SetupFileCatalog, StationCatalog, \
                    FrequencyCatalog
from . import parameter, setbbc

import schedlib as s

from collections import defaultdict

def setdefs():
    station_catalog = StationCatalog()
    station_catalog.read() 
    station_entries = station_catalog.entries
    
    setup_catalog = SetupCatalog()
    setup_catalog.read()
    setup_entries_map = defaultdict(list)
    for entry in setup_catalog.scheduled():
        setup_entries_map[entry.isetnum].append(entry)
    
    setup_file_catalog = SetupFileCatalog()
    setup_file_catalog.read()
    for index, setup_file_entry in enumerate(setup_file_catalog.scheduled(), 1):
        if index in setup_entries_map:
            setup_file_entry.mschn = max(
                setup_entry.nchan 
                for setup_entry in setup_entries_map[index])
            setup_file_entry.minbbc = min(
                station_entries[setup_entry.isetsta-1].nbbc
                for setup_entry in setup_entries_map[index])
        else:
            setup_file_entry.mschn = 0
            setup_file_entry.minbbc = 99
    setup_file_catalog.write(range(len(setup_file_catalog.scheduled())))

    s.setfirm()

    if not s.schn1.vlaonly:
        for index, entry in enumerate(setup_catalog.scheduled(), 1):
            s.setchan(index)
            entry.needcat = s.setfreq(index)
            s.sethw1(index)

    for index, entry in enumerate(setup_catalog.scheduled(), 1):
        s.setfcat(index, entry.needcat)

    setup_catalog.read()
    for index in range(1, len(setup_catalog.scheduled()) + 1):
        setbbc(index, setup_catalog, FrequencyCatalog().read(), 
               station_entries)

    s.setrec()

    setup_catalog.read()
    for index, entry in enumerate(setup_catalog.scheduled(), 1):
        if (entry.track[0, 0] == 0) and \
           any(entry.format.startswith(f) 
               for f in ("VLBA","MKIV", "VDIF", "MARK5B", "LBA")):
            # transpose the track order as settrk returns it in fortran order
            entry.track = s.settrk(
                entry.nchan, entry.tapemode, entry.format, entry.bits,
                entry.bbc, entry.sidebd, entry.dbe, 
                station_entries[entry.isetsta-1].dar, index, 
                s.schn5.twohead, s.schcon.debug, parameter.ilog).T
    setup_catalog.write(range(len(setup_catalog.scheduled())))
    
    s.setusyn()
    setup_catalog.read()
    for entry in setup_catalog.scheduled():
        entry.vfesyn.fill(0)
    setup_catalog.write(range(len(setup_catalog.scheduled())))



from . import parameter, stread
from ..catalog import StationCatalog
from .. import key, util

import schedlib as s

last_station_file = ""

def getsta(stdin, values, index, gotvex, mjd1):
    global last_station_file

    maxsta = StationCatalog.maxcat

    if s.schcon.debug and (index < 2):
        s.wlog(1, "GETSTA: starting.")
    if index == 0:
        last_station_file = ""

    station_file = util.expand_file_name(values["stafile"])
    s.schcst.stafile = util.resize_string(station_file, 
                                          s.schcst.stafile.itemsize, "stafile")
    s.schcst.locafile = util.resize_string(util.expand_file_name(
        values["locfile"]), s.schcst.locafile.itemsize, "locfile")
    if (station_file.upper() != "NONE") and (station_file != last_station_file):
        try:
            f = open(station_file, "r")
        except Exception as e:
            s.putout(str(e))
            s.error("Station catalog: {} not opened".format(station_file))
        else:
            with f:
                input_iterator = key.KeyfileLister(f)
                stread(input_iterator, stdin, mjd1)
        
    last_station_file = station_file

    if (index == 0) and (len(values["stations"]) == 0):
        s.errlog("GETSTA: No stations specified!")
    elif len(values["stations"]) == 0:
        s.schn2a.stascn[index] = s.schn2a.stascn[index-1]
    else:
        s.schn2a.stascn[index] = False
        for station in values["stations"]:
            ksta, ista, doit = s.stano(station)
            if doit:
                if ksta == 0:
                    s.wlog(1, "GETSTA: Station {} not found in catalogs".\
                           format(station))
                    s.errlog("GETSTA: Note -- specify station catalog "
                             "before end of first scan input")

                if ista == 0:
                    s.schn1.nsta += 1
                    if s.schn1.nsta > maxsta:
                        s.errlog("'SCHIN: Too many stations!")
                    ista = s.schn1.nsta
                    s.schc1.staname[ista-1] = s.schcst.station[ksta-1]

                    s.schn1.stanum[ista-1] = ksta
                    s.schsta.ischsta[ksta-1] = ista

                    control = bytes(s.schcst.control[s.schn1.stanum[ista-1]]).\
                       decode().strip()
                    if control == "VEX ":
                        gotvex = True

                    if control == "VSOP":
                        s.schcon.dovsop = True

                    if s.schc1.staname[ista-1][:4] != b"VLBA":
                        s.schn5.allvlba = False
                    else:
                        s.schn1.gotvlba = True

                s.schn2a.stascn[index][ista-1] = True
                            
    return gotvex



import schedlib as s

import numpy as np

import pkg_resources

import pickle
import collections
import copy
import re
import itertools
import os.path
import pkg_resources

class KeyinError(RuntimeError):
    pass

subscript_re = re.compile(r"(?P<key>[A-Za-z0-9]+)\((?P<index>\d+)\)")
def merge_record_with_defaults(record, record_defaults, state_defaults):
    """
    Merge record values from keyin reader with defaults,
    state_defaults are updated if new values are in record,
    record_defaults and state_defaults are dicts:
     {key: [default value, operator]}
    key is assumed to be lower case
    returns a (dict for the record merged with default, 
               set of expanded keys present in record) tuple
    where the operator is applied to the supplied value for the key in record
    """

    # There is an opportunity to reduce runtime here:
    # By preprocess defaults with the functions that process keyin values.
    # Right now that is executed on both given values and defaults.
    # Usually there are (way) more defaults than given values.
    # So far, this doesn't seem to be a significant contribution to runtime,
    # so I have chosen to keep things simple for now.

    assert(not (set(record_defaults.keys()) & set(state_defaults.keys())))
    record = {key.lower(): value for key, value in record.items()}
    # keys in record might be starting substring of keys in defaults
    # as long as they are unique, update the keys to their full name
    all_keys = sorted(itertools.chain(record_defaults.keys(),
                                      state_defaults.keys()))
    # keyin allows syntax like '<key>(<n>) = value'
    # store the key part in record keys
    record_keys = []
    # and the n part in start indices
    start_indices = []
    original_key = []
    for key in record.keys():
        match = subscript_re.match(key)
        if match:
            new_key = match.group("key")
            record_keys.append(new_key)
            start_indices.append(int(match.group("index")) - 1)
        else:
            record_keys.append(key)
            start_indices.append(0)
        
    key_indices = np.searchsorted(all_keys, record_keys, side="left")
    values = {key: copy.deepcopy(value[0]) 
              for key, value in record_defaults.items()}
    values.update({key: copy.deepcopy(value[0]) 
                   for key, value in state_defaults.items()})
    try:
        for key, value, key_index, start_index in zip(
                record_keys, record.values(), key_indices, start_indices):
            if (key_index >= len(all_keys)) or \
               (not all_keys[key_index].startswith(key)):
                raise KeyinError("Unknown parameter: {}".format(key))
            if (all_keys[key_index] != key) and \
               (key_index + 1 < len(all_keys)) and \
               (all_keys[key_index+1].startswith(key)):
                raise KeyinError("'Ambiguous keyword: {}".format(key))
            full_key = all_keys[key_index]
            
            # if the default is a list, slice the given value into the default
            if isinstance(values[full_key], list):
                if isinstance(value, str):
                    # don't want to slice strings
                    values[full_key][start_index:start_index+1] = [value]
                else:
                    # assume we do want to slice any other type that supports it
                    try:
                        values[full_key][start_index:start_index+len(value)] = \
                            value
                    except TypeError:
                        values[full_key][start_index:start_index+1] = [value]
            else:
                values[full_key] = value

        # now apply the operator and update the state defaults
        for key, (dummy, operator) in record_defaults.items():
            values[key] = operator(values[key])
        for key, (dummy, operator) in state_defaults.items():
            state_defaults[key][0] = values[key]
            values[key] = operator(values[key])
    except KeyinError:
        raise
    except Exception as e:
        raise KeyinError("Error at key '{}' value '{}'".format(
            key, values.get(key)))

    return (values, set(all_keys[index] for index in key_indices))

# keyin value processing functions
noop = lambda x: x
upper = lambda x: x.upper()
lower = lambda x: x.lower()
foreach = lambda f: (lambda x: [f(e) for e in x])
strip_upper = lambda x: x.upper().strip()
multiply_by = lambda m: (lambda x: x * m)
extend_to = lambda l: (lambda x: x + [x[0]] * (l - len(x)))
# map int to bool (as KEYIN doesn't do bool), 
# the unconventional True if 0 else False comes from SCHED
to_bool = lambda x: x == 0 
def chain(*functions):
    def ret(x):
        for f in functions:
            x = f(x)
        return x
    return ret

def f2str(f_str):
    return bytes(f_str).decode().rstrip()

def f2str_array(a):
    # arrays of string are actually returned as matrices of numpy.bytes_
    # with a fotran/C mix-up in memory layout
    # eg: ["foo", "bar", "", ""] is returned as
    #     [["f", "a", ""], ["o", "r", ""], ["o", "", ""], ["b", "", ""]]
    return np.array(["".join(c.decode() for c in row).rstrip()
                     for row in a.flatten(order="F").reshape(a.shape)], 
                    dtype=object)

def expand_file_name(shell_file_name):
    return os.path.expanduser(os.path.expandvars(shell_file_name))

def get_catalog_dir():
    if "SCHED" in os.environ:
        return os.path.join(os.environ["SCHED"], "catalogs")
    else:
        return os.path.join(pkg_resources.resource_filename("pysched", ".."), 
                            "catalogs")

bool2str = lambda b: "T" if b else "F"

def resize_string(text, size, attr):
    ret = text.ljust(size)
    if len(ret) > size:
        ret = ret[:size]
        s.wlog(1, ("Warning, maximum string length exceeded for {attr}, "
                   "truncating '{text}' to '{ret}'").format(
                       attr=attr, text=text, ret=ret))
    return ret



from . import scndup
from ..catalog import ScanCatalog, StationCatalog
from ..util import f2str

import schedlib as s

import numpy as np

scan_catalog = ScanCatalog()
station_catalog = StationCatalog()
def makescn(last_scan_index, scan_index, j_scan, source_index, source_name, 
            approx_time, min_elevation, use_time):
    if s.schcon.debug:
        s.wlog(0, "MAKESCN starting.")

    scndup(scan_index - 1, j_scan - 1, False, "MAKESCN")
    
    scans = scan_catalog.direct_access_entries
    scans[scan_index - 1].srcnum = source_index
    scans[scan_index - 1].scnsrc = source_name.ljust(s.schc2b.scnsrc.itemsize)

    stations = station_catalog.used(use_direct_access=True)

    n_good = 0
    ok_sta = np.full(fill_value=False, dtype=bool, shape=(len(stations),))
    for i, station in enumerate(stations):
        if station.stascn[j_scan - 1]:
            last_time, available_time = s.stageo(scan_index, i + 1, approx_time,
                                                 last_scan_index[i], "MAKESCN")
            if (f2str(station.up1[scan_index - 1]) == "") and \
               (f2str(station.up2[scan_index - 1]) == "") and \
               (station.el1[scan_index - 1] > min_elevation) and \
               (station.el2[scan_index - 1] > min_elevation):
                ok_sta[i] = True
                n_good += 1
    
    for i, station in enumerate(stations):
        station.stascn[scan_index - 1] = station.stascn[scan_index - 1] and \
                                         ok_sta[i]

    if n_good >= 1:
        scans[scan_index - 1].startj = approx_time

        s.opttim(last_scan_index, last_scan_index, scan_index, True, use_time, 
                 False)
        n_good = s.scngeo(last_scan_index, scan_index)

    return n_good



import schedlib as s

import numpy as np

def bbcalt(ks, setup_entry, ifbbc, ifnam, warning, caller):
    if s.schcon.debug:
        s.wlog(0, "BBCALT starting. Called by {}".format(caller))

    max_bbc = ifbbc.shape[0]
    max_if = ifbbc.shape[1]
    ubbc = np.full(max_bbc, False)
    for bbc in setup_entry.bbc:
        if bbc > 0:
            ubbc[bbc - 1] = True
    ifinput = {}

    for ich, channel in enumerate(setup_entry.channel):
        if channel.bbc == 0:
            for other in setup_entry.channel[:ich]:
                if channel.freqref == other.freqref:
                    if channel.ifchan == other.ifchan:
                        channel.bbc = other.bbc
                        break
                    elif channel.altifc == other.ifchan:
                        channel.bbc = other.bbc
                        channel.ifchan = other.altifc
                        break
            
        for ibbc in np.where(np.logical_not(ubbc))[0]:
            ifs = np.where(ifbbc[ibbc] == 1)[0]
            for iif in ifs:
                if channel.bbc != 0:
                    break
                chknam = ifnam[iif]
                if (warning == "DBBC") and (iif in ifinput):
                    chknam = ifinput[iif]

                ifmatch = channel.ifchan.startswith(chknam)
                altifmatch = channel.altifc.startswith(chknam)
                if ifmatch or altifmatch:
                    if not ifmatch:
                        channel.ifchan = channel.altifc
                    channel.bbc = ibbc + 1
                    ubbc[ibbc] = True
                    ifinput[iif] = channel.ifchan
                    
        if channel.bbc == 0:
            s.wlog(1, " ")
            s.wlog(1, "BBCALT:  BBC setting problem: ")
            s.wlog(1, "         Setup file: {}".format(setup_entry.setname))
            s.wlog(1, "         Station: {}".format(setup_entry.setsta[0]))
            s.wlog(0, "         ICHAN={}  IFNAME={}  ALTIFC={}  NNBBC={}".\
                   format(ich+1, channel.ifchan, channel.altifc, max_bbc))
            s.wrtmsg(0, "BBCALT", "noassignbbc")
            if ich > 4:
                s.wlog(1, "         Maybe too many channels are requested for "
                       "the same IF channel.")
                
            if warning in ("GEO1", "GEO2"):
                s.wlog(0, "Note that 'geodetic' VLBA systems have restrictions "
                       "on IF assignments.")
                s.wlog(0, "        See IFCHAN in the manual.")
            if warning == "GEO2":
                s.wlog(0, "  They also cannot use 2 bits from more than "
                       "8 BBCs.")
            elif warning == "S2":
                s.wlog(0, "Note that the S2 system is not hooked to all BBCs.")
            
            s.errset(ks)



import schedlib as s

import numpy as np

def sttant(tantsta1, tantsta2):
    s.schn1.tants1[:s.schn1.nsta] = False
    s.schn1.tants2[:s.schn1.nsta] = False

    tantsta1 = set(tant.ljust(s.schc1.staname.itemsize).encode() 
                   for tant in tantsta1)
    tantsta2 = set(tant.ljust(s.schc1.staname.itemsize).encode() 
                   for tant in tantsta2)
    
    s.schn1.tants1 = [station in tantsta1 for station in s.schc1.staname]
    s.schn1.tants2 = [station in tantsta2 for station in s.schc1.staname]

    tant1or2 = np.logical_or(s.schn1.tants1, s.schn1.tants2)
    control = set(("SNAP", "SN50", "VEX", "NRAO"))
    for index in range(s.schn1.nsta):
        if (s.schn1.tants1[index] or s.schn1.tants2[index]) and \
           (s.schcst.control[s.schn1.stanum[index]-1].decode().strip() in 
            control):
            s.wlog(1, "SCHIN: TANT request for {} will be ignored.".format(
                s.schcst.station[s.schn1.stanum[index]-1].decode().strip()))
    if not (s.schn1.tants1[:s.schn1.nsta].any() or \
            s.schn1.tants2[:s.schn1.nsta].any()):
        mask = (s.schc1.staname[:s.schn1.nsta] == 
                "EFLSBERG".ljust(s.schc1.staname.itemsize).encode())
        s.schn1.tants1[:s.schn1.nsta][mask] = True



from .catalog import Catalog

import schedlib as s

class LineRestFrequencyCatalog(Catalog):
    """
    INTEGER          NRESTFQ(MAXLGP)
    DOUBLE PRECISION RESTFREQ(MAXLCH,MAXLGP)
    CHARACTER        LINENAME(MAXLGP)*8
    """

    maxlgp = s.schlin.nrestfq.shape[0]
    block_items = {
        s.schlin : [
            "restfreq",
            "nrestfq" # according to SCHED comment, not used
        ],
        s.schcli : [
            "linename"
        ]
    }

    def __init__(self):
        super().__init__(self.maxlgp, self.block_items)




from .catalog import Catalog, get_arrays, write, map_attr_array
from ..util import f2str

import schedlib as s

import numpy as np

class StationCatalog(Catalog):
    """
    INTEGER          ISCHSTA(MAXCAT), MJDRATE(MAXCAT)
    CHARACTER        STATION(MAXCAT)*8
    CHARACTER        STCODE(MAXCAT)*3
    CHARACTER        STCODEU(MAXCAT)*3
    DOUBLE PRECISION XPOS(MAXCAT), YPOS(MAXCAT), ZPOS(MAXCAT)
    DOUBLE PRECISION DXPOS(MAXCAT), DYPOS(MAXCAT), DZPOS(MAXCAT)
    DOUBLE PRECISION ELEV(MAXCAT), LAT(MAXCAT), LONG(MAXCAT)
    CHARACTER        POSREF(MAXCAT)*80
    CHARACTER        CONTROL(MAXCAT)*4, MOUNT(MAXCAT)*5
    CHARACTER        DAR(MAXCAT)*5, RECORDER(MAXCAT)*6
    CHARACTER        DISK(MAXCAT)*6, MEDIADEF(MAXCAT)*6
    CHARACTER        TSCAL(MAXCAT)*4, DBBCVER(MAXCAT)*8
    INTEGER          NBBC(MAXCAT), STNDRIV(MAXCAT), NHEADS(MAXCAT)
    LOGICAL          VLBADAR(MAXCAT), USEONSRC(MAXCAT)
    INTEGER          NHORIZ(MAXCAT)
    REAL             HORAZ(200,MAXCAT), HOREL(200,MAXCAT)
    INTEGER          NAXLIM(MAXCAT)
    REAL             AX1LIM(6,MAXCAT), AX2LIM(6,MAXCAT)
    REAL             AX1RATE(MAXCAT), AX2RATE(MAXCAT)
    REAL             AX1ACC(2,MAXCAT), AX2ACC(2,MAXCAT)
    REAL             TSETTLE(MAXCAT), MINSETUP(MAXCAT)
    REAL             MAXSRCHR(MAXCAT), TLEVSET(MAXCAT)
    REAL             ZALIM(MAXCAT), AXOFF(MAXCAT)
    """

    maxcat = s.schsta.ischsta.shape[0]
    block_items = {
        s.schcst: [
            'control', 
            'dar', 
            'dbbcver', 
            'disk', 
            'mediadef', 
            'mount', 
            'posref', 
            'recorder', 
            'station', 
            'stcode', 
            'stcodeu', 
            'tscal'],
        s.schsta: [
            'ax1acc', 
            'ax1lim', 
            'ax1rate', 
            'ax2acc', 
            'ax2lim', 
            'ax2rate', 
            'axoff', 
            'dxpos', 
            'dypos', 
            'dzpos', 
            'elev', 
            'horaz', 
            'horel', 
            'ischsta', 
            'lat', 
            'long_bn', 
            'maxsrchr', 
            'minsetup', 
            'mjdrate', 
            'naxlim', 
            'nbbc', 
            'nheads', 
            'nhoriz', 
            'stndriv', 
            'tlevset', 
            'tsettle', 
            'useonsrc', 
            'vlbadar', 
            'xpos', 
            'ypos', 
            'zalim', 
            'zpos']
    }

    """
    INTEGER          NSETUP(MAXSCN,MAXSTA)
    INTEGER          FSETI(MAXSCN,MAXSTA)
    LOGICAL          STASCN(MAXSCN,MAXSTA)
    LOGICAL          USETAPE(MAXSTA), USEDISK(MAXSTA)
    DOUBLE PRECISION TPSTART(MAXSCN,MAXSTA)
    REAL             GBYTES(MAXSCN,MAXSTA)
    DOUBLE PRECISION LST1(MAXSCN,MAXSTA), LST2(MAXSCN,MAXSTA)
    DOUBLE PRECISION TONSRC(MAXSCN,MAXSTA), TSLEW(MAXSCN,MAXSTA)
    REAL             EL1(MAXSCN,MAXSTA), AZ1(MAXSCN,MAXSTA)
    REAL             HA1(MAXSCN,MAXSTA), PA1(MAXSCN,MAXSTA)
    REAL             EL2(MAXSCN,MAXSTA), AZ2(MAXSCN,MAXSTA)
    REAL             HA2(MAXSCN,MAXSTA), PA2(MAXSCN,MAXSTA)
    CHARACTER        UP1(MAXSCN,MAXSTA)*1, UP2(MAXSCN,MAXSTA)*1
C   from schpeak.inc, MPKSTA is defined as MAXSTA
    INTEGER          PKGROUP(MPKSTA)
    """
    maxsta = s.schn5.usedisk.shape[0]
    scheduled_station_items = {
        s.schn2a: [
            'nsetup',
            'fseti',
            'stascn'],
        s.schn5: [
            'usetape',
            'usedisk',
            'tpstart',
            'gbytes'],
        s.schn6: [
            'lst1',
            'lst2',
            'tonsrc',
            'tslew',
            'el1',
            'az1',
            'ha1',
            'pa1',
            'el2',
            'az2',
            'ha2',
            'pa2'],
        s.schc6: [
            'up1',
            'up2'],
        s.schpeakn: [
            'pkgroup']
    }
    
    # all scheduled scan related attributes have an array shape of
    # (scan, station)
    scan_station_items = {
        block: [attr for attr in attrs if len(getattr(block, attr).shape) == 2]
        for block, attrs in scheduled_station_items.items()}
    non_scan_station_items = {
        block: [attr for attr in attrs if len(getattr(block, attr).shape) == 1]
        for block, attrs in scheduled_station_items.items()}
    
    class DirectAccessCatalogEntry(object):
        # a catalog entry which maps setattr and getattr directly to
        # the COMMON blocks in schedlib

        def __init__(self, index, attr_array, scheduled_attr_array):
            self.__dict__["_index"] = index
            self.__dict__["_attr_array"] = attr_array
            self.__dict__["_scheduled_attr_array"] = scheduled_attr_array

        def __getattr__(self, attr):
            try:
                array = self._attr_array[attr]
                index = self._index
            except KeyError:
                # try the scheduled attributes next
                try:
                    array = self._scheduled_attr_array[attr]
                    # map the index into all (catalog) stations to 
                    # an index into scheduled stations
                    index = next(i for i, e in enumerate(
                        s.schn1.stanum[:s.schn1.nsta])
                                 if e == self._index + 1)
                except KeyError:
                    raise AttributeError("'{}' object has no attribute '{}'".\
                                         format(type(self).__name__, attr))
                except StopIteration:
                    raise RuntimeError(
                        "Station {} has no scheduled attributes.".format(
                            self.station))

            if len(array.shape) > 1:
                return array[..., index]
            else:
                ret = array[index]                    
                if array.dtype.kind == "S":
                    return f2str(ret)
                return ret

        def __setattr__(self, attr, value):
            try:
                array = self._attr_array[attr]
                index = self._index
            except KeyError:
                try:
                    # try the scheduled attributes next
                    array = self._scheduled_attr_array[attr]
                    index = next(i for i, e in enumerate(s.schn1.stanum)
                                 if e == self._index + 1)
                except KeyError:
                    return super().__setattr__(attr, value)
                except StopIteration:
                    raise RuntimeError(
                        "Station {} has no scheduled attributes.".format(
                            self.station))

            array[..., index] = value


    def __init__(self):
        super().__init__(self.maxcat, self.block_items)
        self.scheduled_attr_array = map_attr_array(self.scheduled_station_items)
        self.non_scan_attr_array = map_attr_array(self.non_scan_station_items)
        self.has_scheduled_attributes_for = []
        self.direct_access_entries = [
            self.DirectAccessCatalogEntry(i, self.attr_array, 
                                          self.scheduled_attr_array) 
            for i in range(self.maxcat)]
    
    def write(self, indices=None):
        if indices is None:
            entries = self.entries
        else:
            entries = [self.entries[i] for i in indices]
        for entry in entries:
            entry.nhoriz = len(entry.horaz)
        return super().write(indices)

    def adjust_lengths(self, entries):
        for entry in entries:
            length = entry.nhoriz
            entry.horaz = entry.horaz[:length]
            entry.horel = entry.horel[:length]
        
    def scheduled_slice(self):
        return slice(s.schsta.msta)

    def used(self, use_direct_access=False):
        entries = self.direct_access_entries if use_direct_access \
                  else self.entries
        return [entries[i-1] for i in s.schn1.stanum[:s.schn1.nsta]]

    def _read_scheduled_attributes_implementation(self, arrays):
        for index, entry in enumerate(self.used()):
            for key, value in arrays.items():
                setattr(entry, key, value[index])
    
    def read_scheduled_attributes(self):
        """
        Pre: self has entries
        """
        arrays = get_arrays(self.scheduled_attr_array)
        self._read_scheduled_attributes_implementation(arrays)
        self.has_scheduled_attributes_for = s.schn1.stanum[:s.schn1.nsta].copy()
    
    def read_scheduled_attributes_for_scan(self, scan_index):
        """
        Pre: self has entries
        Read scan related attributes for @scan_index into station entries.
        """
        if not np.array_equal(self.has_scheduled_attributes_for,
                              s.schn1.stanum[:s.schn1.nsta]):
            # initialize the station entry scan arrays
            self.read_scheduled_attributes()
            return
        
        for block, attrs in self.scan_station_items.items():
            for attr in attrs:
                array = getattr(block, attr)
                for station_index, entry in enumerate(self.used()):
                    value = array[scan_index, station_index]
                    if array.dtype.kind == "S":
                        value = value.decode()
                    getattr(entry, attr)[scan_index] = value

    def read_non_scan_scheduled_attributes(self):
        """
        Pre: self has entries
        Read scheduled attributes not related to scans into station entries.
        """
        arrays = get_arrays(self.non_scan_attr_array)
        self._read_scheduled_attributes_implementation(arrays)

    def write_scheduled_attributes(self):
        """
        Pre: read_scheduled_attributes is up to date
        """
        entries = self.used()
        write(self.scheduled_attr_array, entries, range(len(entries)))
            



from .parameter import secpday, sidereal
from ..catalog import ScanCatalog, StationCatalog, SourceCatalog
from ..util import f2str
from . import scndup

import schedlib as s

import numpy as np

import math

scan_catalog = ScanCatalog()
station_catalog = StationCatalog()

max_ins = ScanCatalog.maxscan // 2
scan_used = np.empty(dtype=int, shape=(max_ins,))
ha_min = np.empty(dtype=float, shape=(max_ins,))
ha_max = np.empty(dtype=float, shape=(max_ins,))
ha_begin = np.empty(dtype=float, shape=(max_ins,))
ha_end = np.empty(dtype=float, shape=(max_ins,))
t_ha_min = np.empty(dtype=float, shape=(max_ins,))
t_ha_max = np.empty(dtype=float, shape=(max_ins,))
t_use = np.empty(dtype=float, shape=(max_ins,))
t_avail = np.empty(dtype=float, shape=(max_ins,))
source_used = np.empty(dtype=int, shape=(max_ins,))
source_look = np.empty(dtype=int, shape=(max_ins,))
scan_weight = np.empty(dtype=float, shape=(max_ins,))
op_ha_t = np.empty(dtype=float, shape=(max_ins,))
op_ha_wid_t = np.empty(dtype=float, shape=(max_ins,))

def opthas(last_scan_index, k_scan, scan_index):
    global n_out, scan_used, dur_total, ref_station_index, \
        ha_min, ha_max, ha_begin, ha_end, t_ha_min, t_ha_max, t_use, t_avail, \
        source_used, source_look, \
        skip_time, skip_inc, n_skip, sd_total, \
        op_ha_t, op_ha_wid_t


    scans = scan_catalog.direct_access_entries
    stations = station_catalog.used(use_direct_access=True)
    
    if k_scan == 1:
        # initialize for multiple calls to this function
        n_out = 0
        skip_time = 0
        n_skip = 0
        skip_inc = 5 * 60 / secpday
        if s.schn1.nscans > max_ins:
            s.errlog("OPHAS: Too many requested scans for OPTMODE=HAS.  Max: {} "
                     "Requested: {}".format(max_ins, s.schn1.nscans))
        if s.schcon.opdur <= 0:
            s.errlog("OPHAS: Please specify OPDUR for OPTMODE=HAS")

        min_duration = min(scan.dur for scan in scans[:s.schn1.nscans])
        if (s.schcon.opdur * sidereal - min_duration) > 1:
            s.errlog("OPHAS: OPTMODE=HAS not meant for observations over "
                     "24 hours sidereal + min scan dur.")

        ht_stop = scans[0].startj + s.schcon.opdur

        _, ref_station_index, _ = s.stano(f2str(s.schsco.ophasta))
        source_catalog = SourceCatalog()
        source_used = np.zeros(shape=(len(source_catalog.entries),), dtype=int)
        source_look = np.zeros(shape=(s.schn1.nscans,), dtype=int)
        dur_total = 0
        sd_total = 0
        for i, scan in enumerate(scans[:s.schn1.nscans]):
            source_index = scan.srcnum
            source_used[source_index - 1] += 1
            source_look[i] = source_used[source_index - 1]
            scan_used[i] = 0
            dur_total += scan.dur

        if dur_total > s.schcon.opdur:
            s.wlog(1, "OPTHAS: ** The total time in requested scans of "
                   "{} hours".format(dur_total * 24))
            s.wlog(1, "           exceeds the experiment duration (OPDUR) of "
                   "{} hours.".format(s.schcon.opdur * 24))
            s.wlog(1, "           Not all scans can be scheduled.")

        ref_station = stations[ref_station_index - 1]
        for j_scan, scan in enumerate(scans[:s.schn1.nscans]):
            ha_min[j_scan], ha_max[j_scan], t_ha_min[j_scan], t_ha_max[j_scan] \
                = s.halim(j_scan + 1, ref_station_index)
            
            ref_station.ha1[j_scan], ref_station.el1[j_scan], \
                ref_station.az1[j_scan], ref_station.lst1[j_scan], \
                ref_station.pa1[j_scan] = s.schgeo(
                    j_scan + 1, ref_station_index, scans[0].startj)
            ref_station.ha2[j_scan], ref_station.el2[j_scan], \
                ref_station.az2[j_scan], ref_station.lst2[j_scan], \
                ref_station.pa2[j_scan] = s.schgeo(
                    j_scan + 1, ref_station_index, ht_stop - scan.dur)

            ha_begin[j_scan] = ref_station.ha1[j_scan]
            ha_end[j_scan] = ref_station.ha2[j_scan]

            t_avail[j_scan], t_use[j_scan] = s.haavai(
                t_ha_min[j_scan], t_ha_max[j_scan], 
                scans[0].startj, ht_stop - scan.dur,
                source_used[scan.srcnum - 1], source_look[j_scan])
            
            if t_avail[j_scan] == -1:
                s.wlog(1, " ")
                s.wlog(1, " ****** {} is never above OPMINEL and the horizon at "
                       "OPMINANT stations at once.".format(scan.scnsrc))
                s.errlog("Leave out the source or change OPMINANT or OPMINEL")

            if scan.opha == 0:
                op_ha_t[j_scan] = t_use[j_scan]
            else:
                op_ha_t[j_scan] = (scan.opha - ha_begin[j_scan]) \
                                  / sidereal + scans[0].startj
                if op_ha_t[j_scan] < scans[0].startj:
                    op_ha_t[j_scan] +=  24 / sidereal
                    if op_ha_t[j_scan] > scans[0].startj + s.schcon.opdur:
                        s.wlog(1, "OPHAS: Requested hour angle outside range: "
                               "scan, source, OPHA: {} {} {}".format(
                                   j_scan + 1, scan.scnsrc, scan.opha))

            if scan.ophawid == 0:
                op_ha_wid_t[j_scan] = 0.7 * t_avail[j_scan] \
                                      / source_used[scan.srcnum - 1]
            else:
                op_ha_wid_t[j_scan] = scan.ophawid / secpday
    else:
        ref_station = stations[ref_station_index - 1]
        
    if n_out > s.schn1.nscans:
        done = True

    else:
        done = False
        found_scan = False
        while not found_scan:
            if k_scan == 1:
                approx_time = scans[0].startj + skip_time
                adjust = False
            elif skip_time != 0:
                approx_time = scans[scan_index - 2].stopj + skip_time
                adjust = False
                if approx_time > scans[0].startj + s.schcon.opdur:
                    done = True
                    break
            else:
                approx_time = scans[scan_index - 2].stopj
                adjust = True

            weight_max = 0
            weight_max_scan = -1
            if s.schcon.opprtlev >= 3:
                s.wlog(0, "              Scan   Source     New scan time "
                       "Opt scan time  TIMEWT  SLEWWT  RISEWT   SETWT   SCNWT "
                       "    dT      WID  SlewSec")

            for j_scan, scan in enumerate(scans[:s.schn1.nscans]):
                if scan_used[j_scan] == 0:
                    scan_weight[j_scan] = 0
                    ref_station.ha1[j_scan], ref_station.el1[j_scan], \
                        ref_station.az1[j_scan], ref_station.lst1[j_scan], \
                        ref_station.pa1[j_scan] = s.schgeo(
                            j_scan + 1, ref_station_index, approx_time)
                    if (ref_station.ha1[j_scan] >= ha_min[j_scan]) and \
                       (ref_station.ha1[j_scan] <= ha_max[j_scan]):
                        at_source = 0
                        for i, station in enumerate(stations):
                            if station.stascn[j_scan]:
                                _, station.tonsrc[j_scan] = s.stageo(
                                    j_scan + 1, i + 1, approx_time, 
                                    last_scan_index[i], "OPTHAS")
                                if (k_scan > 1) and \
                                   (f2str(station.up1[j_scan]) == ""):
                                    at_source = max(at_source, 
                                                    station.tonsrc[j_scan])

                        if k_scan > 1:
                            op_slew = (at_source - scans[scan_index - 2].stopj) \
                                      * secpday
                        else:
                            op_slew = 0

                        source_separation = 1
                        if k_scan > 1:
                            for last_scan in scans[s.schn1.scan1 - 1:
                                                   s.schn1.scanl]:
                                if scan.scnsrc == last_scan.scnsrc:
                                    time_space = approx_time - last_scan.stopj
                                    source_separation = min(source_separation,
                                                            time_space)
                                    
                        if (source_separation < (scan.opminsep * 
                                                 op_ha_wid_t[j_scan])) or \
                            (abs(approx_time - op_ha_t[j_scan]) > 
                             scan.ophmaxdt / secpday):
                            scan_weight[j_scan] = 0
                        else:
                            time_weight = scan.ophawt * \
                                      (0.5 + (1 / math.pi) * math.atan(
                                          1.5 * (approx_time - op_ha_t[j_scan])
                                          / op_ha_wid_t[j_scan]))

                            slew_weight = 0
                            if (scan.scnsrc != scans[scan_index - 2].scnsrc) \
                               and (scan.opslewti > 0):
                                slew_weight = max(
                                    0, scan.opslewwt * (1 - op_slew / 
                                                        scan.opslewti))

                            rise_weight = 0
                            set_weight = 0
                            if scan.ophlimti > 0:
                                rise_weight = max(
                                    0, scan.ophlimwt * (1 - abs(
                                        ref_station.ha1[j_scan] - ha_min[j_scan])
                                                        * 3600 / scan.ophlimti))
                                set_weight = max(
                                    0, scan.ophlimwt * (1 - abs(
                                        ref_station.ha1[j_scan] - ha_max[j_scan])
                                                        * 3600 / scan.ophlimti))

                            scan_weight[j_scan] = time_weight + slew_weight \
                                                  + rise_weight + set_weight
                            if scan_weight[j_scan] > weight_max:
                                weight_max = scan_weight[j_scan]
                                weight_max_scan = j_scan

                            if s.schcon.opprtlev >= 3:
                                year, day1, time_rad = s.timej(approx_time)
                                start_time_text = f2str(s.tformwrp(
                                    time_rad, "T", 0, 2, 2, "::@"))
                                year, day2, time_rad = s.timej(op_ha_t[j_scan])
                                opt_time_text = f2str(s.tformwrp(
                                    time_rad, "T", 0, 2, 2, "::@"))
                                s.wlog(0, " SCAN WEIGHT {:>5d} {:<12} {:>4d} {} "
                                       "{:>4d} {} {:>7.2f} {:>7.2f} {:>7.2f} "
                                       "{:>7.2f} {:>7.2f} {:>7.2f} {:>7.2f} "
                                       "{:>7.2f}".format(
                                           j_scan + 1, scan.scnsrc, 
                                           day1, start_time_text, 
                                           day2, opt_time_text, 
                                           time_weight, slew_weight, 
                                           rise_weight, set_weight, 
                                           scan_weight[j_scan], 
                                           (approx_time - op_ha_t[j_scan]) * 24, 
                                           op_ha_wid_t[j_scan] * 24, op_slew))

            if weight_max_scan >= 0:
                n_out += 1
                scndup(scan_index - 1, weight_max_scan, False, "OPTHAS")
                new_scan = scans[scan_index - 1]
                new_scan.startj = approx_time
                new_scan.stopj = new_scan.startj + new_scan.dur
                scan_used[weight_max_scan] = scan_index
                skip_time = 0
                sd_total += new_scan.dur
                
                if s.schcon.opprtlev >= 2:
                    year, day, time_rad = s.timej(approx_time)
                    time_text = f2str(s.tformwrp(time_rad, "T", 0, 2, 2, "::@"))
                    s.wlog(0, "USE SCAN {:>5d} {} {:>5d} {}   HA={:>10.2f}  "
                           "Weight: {:10.3f}".format(
                               weight_max_scan + 1, 
                               scans[weight_max_scan].scnsrc,
                               day, time_text, ref_station.ha1[weight_max_scan],
                               weight_max))
                found_scan = True
            else:
                skip_time += skip_inc
                n_skip += 1
                if s.schcon.opprtlev >= 1:
                    year, day, time_rad = s.timej(approx_time)
                    time_text = f2str(s.tformwrp(time_rad, "T", 0, 2, 2, "::@"))
                    s.wlog(0, "OPTHAS:  Skipping {:>6.2f} min at {} {:>7.2f} "
                           "min after scan {:>5d} because no scans are "
                           "available.".format(skip_inc * secpday / 60,
                                               time_text, 
                                               skip_time * secpday / 60,
                                               scan_index))

        if found_scan and (s.schcon.opdur > 0) and \
           (new_scan.stopj > scans[0].startj + s.schcon.opdur):
            s.wlog(1, "OPTHAS:  Schedule stopped because requested total "
                   "duration reached")
            done = True

    if done:
        s.wlog(0, " ")
        s.wlog(0, "OPTHAS:  HAS mode schedule made.")

        s.wlog(0, " ")
        if s.schcon.opprtlev <= 0:
            s.wlog(0, "         OPPRTLEV={}:  Minimum information printed.".\
                   format(s.schcon.opprtlev))
        elif s.schcon.opprtlev == 1:
            s.wlog(0, "         OPPRTLEV=1:  Include summary of fate of each "
                   "input scan.")
        elif s.schcon.opprtlev == 2:
            s.wlog(0, "         OPPRTLEV=2:  Minimum feedback as scans chosen "
                   "plus fate of input scans.")
        elif s.schcon.opprtlev >= 3:
            s.wlog(0, "         OPPRTLEV={}:  Details of each scan choice plus "
                   "summaries.".format(s.schcon.opprtlev))

        s.wlog(0, "         Requested scans: {}".format(s.schn1.nscans))
        s.wlog(0, "         Scheduled scans: {}".format(n_out))
        s.wlog(0, "         Sum of requested durations: {:>7.3f} hr.".format(
            dur_total * 24))
        s.wlog(0, "         Sum of scheduled durations: {:>7.3f} hr.".format(
            sd_total * 24))

        year, day, time_rad = s.timej(scans[s.schn1.scan1 - 1].startj)
        time_text = f2str(s.tformwrp(time_rad, "T", 0, 2, 2, "::@"))
        s.wlog(0, "         Start day and time: {:>5d} {}".format(
            day, time_text))

        year, day, time_rad = s.timej(scans[s.schn1.scanl - 1].stopj)
        time_text = f2str(s.tformwrp(time_rad, "T", 0, 2, 2, "::@"))
        s.wlog(0, "         Stop day and time and total duration (hr): {:>5d} {}"
               " {:>12.3f}".format(day, time_text, 
                                   (scans[s.schn1.scanl - 1].stopj - 
                                    scans[s.schn1.scan1 - 1].startj) * 24))

        s.wlog(0, "         Requested duration (hours) {:>8.2f}".format(
            s.schcon.opdur * 24))

        s.wlog(0, "         Maximum allowed deviation from optimum time "
               "(hr; scan 1) {:>8.3f}".format(scans[0].ophmaxdt / 3600))

        s.wlog(0, "         Number of {:>7.3f} minutes skips because of no scan "
               "available: {}".format(skip_inc * secpday / 60, n_skip))
        
        s.wlog(0, "         Total time in skips: {:>8.3f} minutes.  "
               "(Some may be after last scan)".format(
                   skip_inc * n_skip * secpday / 60))
        
        if s.schcon.opprtlev >= 1:
            s.wlog(0, " ")
            s.wlog(0, "OPTHAS:  Input scan information.  Columns are:")
            s.wlog(0, "         Output scan number and time ")
            s.wlog(0, "         Reference station hour angle.  "
                   "for scheduled scan.")
            s.wlog(0, "         Allowed reference station hour angle range "
                   "(for scan start).")
            s.wlog(0, "         Reference station hour angles for beginning "
                   "and end of observations.")
            s.wlog(0, "         Total hours this source is available.")
            s.wlog(0, "         The reference station is: {}".format(
                ref_station.station))
            s.wlog(0, "  Input    Source    Output      Time          HA        "
                   "Time        HA      HA      HA      HA  Available     "
                   "Rise         Set")
            s.wlog(0, "   Scan               Scan   Day    hms       "
                   "Scan        Opt       Min     Max    Start    End    Hours")

            for j, scan in enumerate(scans[:s.schn1.nscans]):
                if scan_used[j] != 0:
                    year, day1, time_rad = s.timej(scans[scan_used[j] - 1].startj)
                    start_time_text = f2str(
                        s.tformwrp(time_rad, "T", 0, 2, 2, "::@"))
                    ha_p = ref_station.ha1[scan_used[j] - 1]
                else:
                    day1 = 0
                    start_time_text = " "
                    ha_p = 0

                year, day2, time_rad = s.timej(op_ha_t[j])
                opt_time_text = f2str(s.tformwrp(time_rad, "T", 0, 2, 2, "::@"))
                year, day_rise, time_rad = s.timej(t_ha_min[j])
                rise_time_text = f2str(s.tformwrp(time_rad, "T", 0, 2, 2, "::@"))
                year, day_set, time_rad = s.timej(t_ha_max[j])
                set_time_text = f2str(s.tformwrp(time_rad, "T", 0, 2, 2, "::@"))
                s.wlog(0, "{:>5d}   {:<12} {:>5d} {:>5d}  {:<8} {:>8.3f} {:>4d}"
                       "  {} {:>7.3f} {:>7.3f} {:>7.3f} {:>7.3f} {:>7.3f} {:>4d}"
                       " {} {:>4d} {}".format(
                           j + 1, scan.scnsrc, scan_used[j], day1, 
                           start_time_text, ha_p, day2, opt_time_text, ha_min[j],
                           ha_max[j], ha_begin[j], ha_end[j], t_avail[j] * 24, 
                           day_rise, rise_time_text, day_set, set_time_text))

        s.wlog(1, "OPTHAS: There were {} requested scans and "
               "{} Scheduled scans.".format(s.schn1.nscans, n_out))
        s.wlog(1, "OPTHAS: See the sched.runlog for details.")

    return adjust, True, done
